/* $Id$ */

/* GenLib2 - General Library 2
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef GENLIB2_GRID_INCLUDED
#define GENLIB2_GRID_INCLUDED

#define DB_WARNING 0

#include <cassert>
//#include <stream.h>
#include <cfloat>
#include "ex.h"
//#include "vector2.h"
#include "vector3.h"

using namespace std;


namespace GenLib2
{

template<class T>
class Rect {			// rectangle
public:
	T x1, x2, y1, y2;	// interval endpoints 1 (lower-left) and 2 (upper-right)
};


template<class T>
class Grid {						// implementation of the grid method
public:
	Grid();							// constructor
	Grid(int nr, int nc);			// constructor with number of grid squares
	~Grid();						// destructor

	void init(int nr, int nc);		// initilize
	void clear();					// clear grid
	void free();					// destroy everything
	void set(T xo, T yo, T w);		// set origin and size of grid squares
	void set(Rect<T>);				// set origin and size using range
	void insert(Vector3<T> p);		// insert point
	int size();						// number of points
	void info(ostream& os);			// print grid parameter
	void print(ostream& os);		// print grid points
	void occupied(ostream& os);		// print number of points per grid square
	int search(const Rect<T>& range);	// number of points inside rectangle
	Vector3<T>* closest(const Rect<T>& range, const Vector3<T>& o);
															// closest point
	int average(const Rect<T>& range, Vector3<T>& o);	// average

protected:
	struct node {
		Vector3<T>	p;
		struct node* next;
	};
	struct node*** grid;		// 2D matrix of grid squares
	struct node* z;
	T		_xo, _yo;			// origin, upper-left
	T		_width;				// size of grid square
	int		_nrows, _ncols;		// number of grid squares

private:
	//double min2;
	//struct dpoint cp1, cp2;
	//void check(struct dpoint p1, struct dpoint p2);
	int insiderect(const Vector3<T>& p, const Rect<T>& rect);
};


// implementation

// initilize
// create 2-D grid and initilize all grid nodes with a dummy tail node

template <class T>
void Grid<T>::init(int nr, int nc)
{
	DB_ENTER("Grid::init");
	assert(nr > 0 && nc > 0);

	_nrows = nr;
	_ncols = nc;

	grid = new struct node**[_nrows];
	assert(grid != NULL);

	grid[0] = (struct node**)new struct node*[_ncols * _nrows];
	assert(grid[0] != NULL);

	for (int y = 1; y < _nrows; y++)
		grid[y] = grid[y-1] + _ncols;

	z = new node;
	assert(z != NULL);
	for (int r = 0; r < _nrows; r++)
		for (int c = 0; c < _ncols; c++)
			grid[r][c] = z;
}


// erase all points
// delete all nodes

template<class T>
void Grid<T>::clear()
{
    DB_ENTER("Grid::clear");

	struct node *t, *p;

	for (int r = 0; r < _nrows; r++)
		for (int c = 0; c < _ncols; c++)
			{
			t = grid[r][c];
			while (t != z)
				{
				p = t;
				t = t->next;
				delete p;
				}
			grid[r][c] = z;
			}
}


// delete grid
// delete all nodes and the grid itself

template <class T>
void Grid<T>::free()
{
	DB_ENTER("Grid::free");
	if (grid == NULL) return;
	clear();
	delete[] grid[0];
	delete[] grid;
	grid = NULL;
}


// default constructor, without size

template <class T>
Grid<T>::Grid()
{
	DB_ENTER("Grid::default constr");
	_xo = _yo = 0;
	_nrows = _ncols = 0;
	grid = NULL;
	_width = 0;
}


// constructor, with size

template <class T>
Grid<T>::Grid(int nr, int nc)
{
	DB_ENTER("Grid::default constr");
	_xo = _yo = 0;
	_width = 0;
	init(nr, nc);
}


// destructor

template <class T>
Grid<T>::~Grid()
{
	DB_ENTER("Grid::destructor constr");
	free();
}


// set grid resolution

template <class T>
void Grid<T>::set(T xo, T yo, T w)
{
    DB_ENTER("Grid::set");
	_xo = xo;
	_yo = yo;
	_width = w;
}


// set grid resolution using range

template <class T>
void Grid<T>::set(Rect<T> range)
{
    DB_ENTER("Grid::set");

	T xmin = range.x1 < range.x2 ? range.x1 : range.x2;
	T xmax = range.x1 > range.x2 ? range.x1 : range.x2;

	T ymin = range.y1 < range.y2 ? range.y1 : range.y2;
	T ymax = range.y1 > range.y2 ? range.y1 : range.y2;

	T wx = (xmax - xmin) / _ncols;
	T wy = (ymax - ymin) / _nrows;

	_width = wx > wy ? wx : wy;

	_xo = xmin;
	_yo = ymin + _nrows * _width;
}


// insert point in grid

template <class T>
void Grid<T>::insert(Vector3<T> p)
{
    DB_ENTER("Grid::insert");

	struct node* t = new node;
	assert(t != NULL);

	int r = (_yo - p.y) / _width;
	int c = (p.x - _xo) / _width;

	if (r < 0) r = 0;
	if (c < 0) c = 0;
	if (r >= _nrows) r = _nrows - 1;
	if (c >= _ncols) c = _ncols - 1;
	cout << p.x << " " << p.y << " " << r << " " << c << endl;

	t->p = p;
	t->next = grid[r][c];
	grid[r][c] = t;
}


// count number of points

template <class T>
int Grid<T>::size()
{
	DB_ENTER("Grid::size");
	int i, j, n = 0;
	struct node* t;

	for (i = 0; i < _nrows; i++)
		for (j = 0; j < _ncols; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				n++;

    return n;
}


// print info

template <class T>
void Grid<T>::info(ostream &os)
{
	DB_ENTER("Grid::info");
	os << "Number of grid squares : rows = "
		<< _nrows << "  cols = " << _ncols << endl;
	os << "Origin (upper-left)    : xo = " << _xo << "  yo = " << _yo << endl;
	os << "Size of grid squares   : " << _width << endl;
	os << "Grid range:" << endl;
	T dx = _ncols * _width;
	T dy = _nrows * _width;
	os << "  x min = " << _xo << "  x max = " << _xo + dx
		<< "  diff = " << dx << endl;
	os << "  y min = " << _yo - dy << "  y max = " << _yo
		<< "  diff = " << dy << endl;
}


// print

template <class T>
void Grid<T>::print(ostream& os)
{
	DB_ENTER("Grid::print");
	int i, j;
	struct node* t;

	for (i = 0; i < _nrows; i++)
		for (j = 0; j < _ncols; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				os << t->p << endl;
}


// print number of points per grid square

template <class T>
void Grid<T>::occupied(ostream& os)
{
	DB_ENTER("Grid::occupied");
	int i, j;
	struct node* t;

	for (i = 0; i < _nrows; i++)
		{
		os << "row " << i << ": ";
		for (j = 0; j < _ncols; j++)
			{
			int n = 0;
			for (t = grid[i][j]; t != z; t = t->next)
				n++;
			os << n << " ";
			}
		os << endl;
		}
}


// counts number of points inside rectangle

template <class T>
int Grid<T>::search(const Rect<T>& range)
{
	DB_ENTER("Grid::search");
	register struct node *t;
	register int i, j;
	int count = 0;

	int i1 = (_yo - range.y2) / _width;
	if (i1 < 0) i1 = 0;
	int i2 = (_yo - range.y1) / _width;
	if (i2 >= _nrows) i2 = _nrows - 1;
	int j1 = (range.x1 - _xo) / _width;
	if (j1 < 0) j1 = 0;
	int j2 = (range.x2 - _xo) / _width;
	if (j2 >= _ncols) j2 = _ncols - 1;

	for (i = i1; i <= i2; i++)
		for (j = j1; j <= j2; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				if (insiderect(t->p, range))
					count++;

	return count;
}


// find closest point inside range

template <class T>
Vector3<T>* Grid<T>::closest(const Rect<T>& range, const Vector3<T>& o)
{
	DB_ENTER("Grid::closest");
	register struct node *t;
	register int i, j;
	Vector3<T>* c = NULL;
	double dx, dy, ssq, ssq_min = DBL_MAX;

	int i1 = (_yo - range.y2) / _width;
	if (i1 < 0) i1 = 0;
	int i2 = (_yo - range.y1) / _width;
	if (i2 >= _nrows) i2 = _nrows - 1;
	int j1 = (range.x1 - _xo) / _width;
	if (j1 < 0) j1 = 0;
	int j2 = (range.x2 - _xo) / _width;
	if (j2 >= _ncols) j2 = _ncols - 1;

	for (i = i1; i <= i2; i++)
		for (j = j1; j <= j2; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				if (insiderect(t->p, range))
					{
					dx = t->p.x - o.x;
					dy = t->p.y - o.y;
					ssq = dx * dx + dy * dy;
					if (ssq < ssq_min)
						{
						ssq_min = ssq;
						c = &(t->p);
						}
					}

	return c;
}


// calculate average inside the given rectangle by weighted distance (p=1/s^2)
// distance s^2 = dx^2 + dy^2

template <class T>
int Grid<T>::average(const Rect<T>& range, Vector3<T>& o)
{
	DB_ENTER("Grid::average");
	register struct node *t;
	register int i, j;
	register double dx, dy, s2, p;
	register double sum_pz = 0.;
	register double sum_p = 0.;
	register int count = 0;

	int i1 = (_yo - range.y2) / _width;
	if (i1 < 0) i1 = 0;
	int i2 = (_yo - range.y1) / _width;
	if (i2 >= _nrows) i2 = _nrows - 1;
	int j1 = (range.x1 - _xo) / _width;
	if (j1 < 0) j1 = 0;
	int j2 = (range.x2 - _xo) / _width;
	if (j2 >= _ncols) j2 = _ncols - 1;

	for (i = i1; i <= i2; i++)
		for (j = j1; j <= j2; j++)
			for (t = grid[i][j]; t != z; t = t->next)
				if (insiderect(t->p, range))
					{
					dx = t->p.x - o.x;
					dy = t->p.y - o.y;
					s2 = dx * dx + dy * dy;
					if (s2 < DBL_EPSILON) s2 = DBL_EPSILON;
					p = 1. / s2;
					sum_pz += p * t->p.z;
					sum_p += p;
					count++;
					}

	if (sum_p > 0.)
		o.z = sum_pz / sum_p;

	return count;
}


// compute distance and check if less than min
/*
void DGrid::check(struct dpoint p1, struct dpoint p2)
{
	double d2;

	d2 = (p2.x - p1.x) * (p2.x - p1.x) + (p2.y - p1.y) * (p2.y - p1.y);
	if (d2 < min2)
		{
		min2 = d2;
		cp1 = p1;
		cp2 = p2;
		}
}
*/


// check if inside rectangle

template <class T>
inline int Grid<T>::insiderect(const Vector3<T>& p, const Rect<T>& rect)
{
	if (p.x < rect.x1 || p.x > rect.x2 || p.y < rect.y1 || p.y > rect.y2)
		return 0;
	else
		return 1;
}

} // namespace GenLib2

#endif
