/* $Id: matrix.h,v 1.6 2006/01/18 12:01:33 cvsroot Exp $ */

/* GenLib2 - General Library 2
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef GENLIB2_MATRIX_INCLUDED
#define GENLIB2_MATRIX_INCLUDED

#include <cassert>
#include <iostream>
#include <cmath>
using namespace std;

// forward declarations
//template<class T> class IPL::Image;

namespace GenLib2
{

/** Matrix class.
 *  General matrix class.
 *  @author Dirk Stallmann
 */
template <class T>
class Matrix
{
	//friend ostream& operator<<(ostream& os, const Matrix<T>& m);
	//friend class Image<T>;
	
protected:
	/**@name internal helper functions.
	 */
	//@{
		/// allocate memory
		void allocate(int m, int n);

		/// free memory
		void free();

		/// copy array
		void copy(const T* v);

		/// set all elements to value
		void set(const T& val);
	//@}
	
	//T* data() const { return _data; }
	//T* row() const { return _row; }

public:
	/**@name constructors.
	 */
	//@{
		/// constructor without size
		Matrix();

		/// constructor with size and initilize with value
		Matrix(int m, int n, const T& val=T(0));

		/// constructor, initialize with array
		Matrix(int m, int n, const T* v);

		/// constructor of row vector and initilize with zeros
		Matrix(int m, const T& val=T(0));

		/// constructor of row vector, initialize with array
		Matrix(int m, const T* v);

		// with argument list
		//Matrix(int m, int n, T val0, ...);
	
		/// copy constructor
		Matrix(const Matrix& A);
	//@}
	
	/// destructor
	~Matrix();

	/**@name methods.
	 */
	//@{ 
		/// copy operator
		Matrix& operator=(const Matrix& A);

		/// copy sub-matrix A into *this matrix
		Matrix& copy(int rl, int cl, const Matrix& A);

		/// new size
		Matrix& newsize(int m, int n=1);

		/// set all elements to zero
		Matrix& clear() { set(T(0)); return *this; }

		/// set all elements to zero
		void null();

		/// set diagonal elements to 1 (and don't change off-diagonal elements)
		void set_diagonal();

		/** set identity matrix.
		 *  Set diagonal elements to value and non-diagonal elements
		 *  will be cleared.
		 */
		void identity(T val=T(1));

		/// set all elements to value
		Matrix& set_all(const T& val);

		// get and set rows and columns
		// gets and sets rows and columns in (*this). The supplied
		// row/column vector must have exactly the right number of
		// elements. It is not important whether it is an row (1xN)
		// or column (Nx1) vector.
		//void getRow(int rowIndex, Matrix &row);
		void get_col(int c, Matrix& col);
		//void setRow(int rowIndex, const Matrix &row);
		//void setCol(int colIndex, const Matrix &col);

		/// make a square matrix symmetric (by mirroring the upper part)
		void make_upper_symmetric();

		/// make a square matrix symmetric (by mirroring the lower part)
		void make_lower_symmetric();
	//@}
	
	/**@name member data access.
	 */
	//@{
		/// number of rows
		int rows() const { return _m; }
		
		/// number of columns
		int cols() const { return _n; }
		
		/// number of matrix elements
		int size() const { return _m * _n; }

		/// data check
		bool is_data() const { return (_data != NULL); }
		
		/// data check
		bool is_row() const { return (_row != NULL); }

		/// get matrix element function using (i,j) indexing
		T& elem(int i, int j) const { return _row[i][j]; }
		
		/// set matrix element function using (i,j) indexing
		void elem(int i, int j, T val) { _row[i][j] = val; }

		/// get matrix element using (i,j) indexing
		T& operator()(int i, int j) { return _row[i][j]; }
		
		/// get matrix element using (i,j) indexing
		const T& operator()(int i, int j) const { return _row[i][j]; }

		/// get matrix element of row-vector using (i) indexing
		T& operator()(int i) { return _row[i][0]; }
		
		/// get matrix element of row-vector using (i) indexing
		const T& operator()(int i) const { return _row[i][0]; }

		/// get pointer to a row
		T* row(int i) const { return _row[i]; }
		// begin of data
		//T* beg() { return _data[0]; }
		// end of data
		//T* end() { return _data[0] + _m * _n; }

		//int inside(int r, int c) {			// check if inside matrix
		//	return (r >= 0 && c >= 0 && r < _m && c < _n);
		//}
	//@}
	
	/**@name I/O methods.
	 */
	//@{
		/// read matrix
		void read(std::istream& is);
	
		/// print matrix
		void print(ostream& os) const;

		/// print symbolic
		void print_symb(ostream& os, char zero=' ', char nozero='*') const;

		/// print in PBM-format
		void print_pbm(ostream& os) const;
	//@}
	
	/**@name composite assignment operators.
	 */
	//@{
		/// multiplication with a scalar
		Matrix& operator*=(T s);

		/// division by a scalar
		Matrix& operator/=(T s);
	//@}
	
	/**@name 
	 * these are routines taken from Numerical Recipes in C.
	 * They are only to be used for
	 * REAL, SYMMETRIC matrices (square matrices, of course).
	 */
	 //@{
		/** householder reduction
		 * (numerical recipes, page 373)
		 * householder reduction (to a tridiagonal matrix) for
		 * a real, symmetric matrix
		 * (*this) is replaced by the orthogonal matrix Q
		 *  d returns the diagonal elements
		 *  e returns the off-diagonal elements with e(0)=0
		 */
		void tred2(Matrix& d, Matrix& e);

		/** QL algorithm with implicit shifts
		 * (numerical recipes, page 380)
		 *  d and e are the elements of a real, symmetric, tridiagonal
		 *     matrix (d=diagonal, e=off-diagonal elements, e(0) arbitrary).
		 *  (*this) is either the identity matrix in which case the
		 *    eigenvectors of the tridiagonal matrix are returned or
		 *     the matrix as returned by tred2. Then, the eigenvectors
		 *    of the original 'A'-matrix are returned.
		 *  maxIter is the maximum allowed number of iterations (if present)
		 *
		 * RETURNS:
		 *  d(i)   : eigenvalue i
		 *  e      : destroyed
		 *  (*this): k-th column contains normalized eigenvector
		 *           corresponding to d(i)
		 *  All results are only valid if the routine returns TRUE
		 */
		bool tqli(Matrix& d, Matrix& e, int maxIter=30);

		/**
		 * Determines eigenvalues and eigenvectors f the (*this) matrix
		 * for a real, symmetric matrix
		 *  RETURNS:
		 * (the value of *this is preserved)
		 * eigenVals: Nx1 vector of all eigenvalues
		 * eigenVects: NxN matrix of all eigenvectors, written as columns
		 * (i-th column eigenVects(*,i) corresponds to eigenvalue eigenVals(i))
		 *
		 * Result only valid if the routine returns TRUE
		 */
		bool eigensystem(Matrix& eigen_vals, Matrix& eigen_vecs);
	//@}
	
protected:
	/// matrix dimension, number of rows
	int		_m;
	
	/// matrix dimension, number of columns
	int		_n;

	/// matrix elements
	T*		_data;

	/// row-pointer
	T**		_row;
};


// non-member functions...

// I/O operators

/// input operator
template <class T>
istream& operator>>(istream& is, Matrix<T>& m);

/// output operator
template <class T>
ostream& operator<<(ostream& os, const Matrix<T>& m);


// basic matrix algorithms

/// transpose matrix
template <class T>
Matrix<T> transpose(const Matrix<T>& A);

// add/sub matrix m to/from the *this matrix
//Matrix &add(const Matrix &m);
//Matrix &sub(const Matrix &m);

/// addition operator
template <class T>
Matrix<T> operator+(const Matrix<T>& A, const Matrix<T>& B);

/// subtraction operator
template<class T>
Matrix<T> operator-(const Matrix<T>& A, const Matrix<T>& B);

/// matrix multiplication function using references
template <class T>
void mult(Matrix<T>& C, const Matrix<T>& A, const Matrix<T>& B, bool clr=true);

/// matrix multiplication function
template <class T>
Matrix<T> mult(const Matrix<T>& A, const Matrix<T>& B);

/// matrix multiplication operator
template <class T>
Matrix<T> operator*(const Matrix<T>& A, const Matrix<T>& B);

/// general matrix multipication function C = A * B
// matrix status
//   clr=false ... Matrix C accummulated
//   clr=true  ... Matrix C cleared in advance
//   atr=false ... matrix A as defined
//   atr=true  ... matrix A transposed
//   btr=false ... matrix B as defined
//   btr=true  ... matrix B transposed
template <class T>
void mult_gen(Matrix<T>& C, bool clr, const Matrix<T>& A, bool atr,
	const Matrix<T>& B, bool btr);

/// multiplication with scalar
template <class T>
Matrix<T> operator*(double s, const Matrix<T>& A);


// implementation
#include "matrix.cpp"

} // namespace GenLib2

#endif
